/*
hdbck.c Version 0.2.0  Hash database check
Copyright (C) 2003-2010  dondalah721@yahoo.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

/* return code 0 is good database check */
/* return code not zero is a corrupted database */

#include <bthash.h>

typedef struct stat statfmt;

int hdbck(hshhdr)
hashfmt *hshhdr;
   {
   int i;
   off_t ofst;
   off_t ptr;
   int rcdnum;
   int adjadr;
   int prod;
   int indx;
   int bit;
   int msk;
   int bitnum;
   int rcds;
   int tot;
   int rslt,lockrslt;
   int bitmapsz;
   unsigned char *p,*q;
   unsigned char *bitmap;
   statfmt *stbuf;
   statfmt statbuf;
   hshhdr->reason = 0;
   lockrslt = lockf(hshhdr->hndl,F_TEST,0);
   if (lockrslt < 0)
      {
      return(ISRTLOCK3);
      } /* can't insert when file is locked */

   stbuf = (statfmt *) &statbuf;
   rslt = fstat(hshhdr->hndl, stbuf);
   if (rslt < 0)
      {
      fprintf(stderr,"hdbck: stat error on %s\n",
	 hshhdr->dbname);
      perror("stat error");
      hshhdr->reason = SYS_ERR + errno;
      return(SYS_ERR + errno);
      } /* stat err */

   adjadr = stbuf->st_size - HDRLEN;
   rcds   = adjadr / hshhdr->rcdnode;
   prod   = rcds   * hshhdr->rcdnode;
   if (rcds < hshhdr->totrcds)
      {
      fprintf(stderr,"hdbck: database too small "
	 "for hashing address space.\n");
      fprintf(stderr,"hdbck: number of records "
	 "is %d\n", rcds);
      hshhdr->reason = HASH_ERR + 606;
      return(HASH_ERR + 606);
      } /* invalid #records */
   if (prod != adjadr)
      {
      fprintf(stderr,"hdbck: database size"
	 "not on record boundary.\n");
      hshhdr->reason = HASH_ERR + 607;
      return(HASH_ERR + 607);
      } /* invalid record boundary */

   bitmapsz = (rcds+7)>>3;
   bitmap = (unsigned char *) malloc(bitmapsz+16);
   if (bitmap == NULL)
      {
      fprintf(stderr,"hdbck: out of memory\n");
      hshhdr->reason = HASH_ERR + 601;
      return(HASH_ERR + 601);
      } /* out of mem */
   p = (unsigned char *) bitmap;
   q = (unsigned char *) p + bitmapsz;
   while (p < q) *p++ = '\0';

   ofst = HDRLEN;
   tot = 0;
   while (tot < rcds)
      {
      ptr = hgetptr(hshhdr,ofst);
      if (hshhdr->reason) return(hshhdr->reason);
      if (ptr && ptr != UINT_MAX)
	 {
         adjadr = ptr - HDRLEN;
         rcdnum = adjadr / hshhdr->rcdnode;
	 prod   = rcdnum * hshhdr->rcdnode;
	 if (prod != adjadr)
	    {
	    fprintf(stderr,"hdbck: "
	       "pointer at %08x not on "
	       "record boundary\n",
	       (unsigned int) ofst);
            hshhdr->reason = HASH_ERR + 608;
            return(HASH_ERR + 608);
	    } /* invalid record boundary */
	 if (rcdnum > rcds)
	    {
	    fprintf(stderr,"hdbck: pointer "
               "at %08x too large\n",
	       (unsigned int) ofst);
            hshhdr->reason = HASH_ERR + 609;
            return(HASH_ERR + 609);
            } /* too large */
         if (rcdnum < hshhdr->totrcds)
	    {
	    fprintf(stderr,"hdbck: pointer "
               "at %08x too small\n",
	       (unsigned int) ofst);
            hshhdr->reason = HASH_ERR + 610;
            return(HASH_ERR + 610);
            } /* too small */
	 indx = rcdnum >> 3;
         p = (unsigned char *) bitmap + indx;
         bit = rcdnum & 7;
         msk = 1 << bit; 
         if (*p & msk)
	    {
	    fprintf(stderr,"hdbck: "
	       "duplicate pointer at %08x\n",
	       (unsigned int) ofst);
            hshhdr->reason = HASH_ERR + 602;
            return(HASH_ERR + 602);
	    } /* dup ptr */
         *p |= msk;
	 } /* if pointer filled */
      ofst += hshhdr->rcdnode;
      tot++;
      } /* for each record in database */

   ptr = hshhdr->freestk;
   if (ptr)
      {
      adjadr = ptr - HDRLEN;
      rcdnum = adjadr / hshhdr->rcdnode;
      prod   = rcdnum * hshhdr->rcdnode;
      if (prod != adjadr)
         {
         fprintf(stderr,"hdbck: "
            "freestk pointer not on "
            "record boundary\n");
         hshhdr->reason = HASH_ERR + 611;
         return(HASH_ERR + 611);
         } /* invalid record boundary */
      if (rcdnum > rcds)
         {
         fprintf(stderr,"hdbck: freestk "
            "pointer too large\n");
         hshhdr->reason = HASH_ERR + 612;
         return(HASH_ERR + 612);
         } /* too large */
      if (rcdnum < hshhdr->totrcds)
         {
         fprintf(stderr,"hdbck: freestk "
            "pointer too small\n");
         hshhdr->reason = HASH_ERR + 613;
         return(HASH_ERR + 613);
         } /* too small */
      indx = rcdnum >> 3;
      p = (unsigned char *) bitmap + indx;
      bit = rcdnum & 7;
      msk = 1 << bit; 
      if (*p & msk)
	 {
	 fprintf(stderr,"hdbck: freestk "
	    "duplicate\n");
         hshhdr->reason = HASH_ERR + 603;
         return(HASH_ERR + 603);
	 } /* dup ptr */
      *p |= msk;
      } /* map the first free block */

   i = bitnum = 0;
   p = (unsigned char *) bitmap;
   while (i < bitmapsz)
      {
      bit = 1;
      while (bit < 255)
	 {
	 bitnum++;
	 if (bitnum <= hshhdr->totrcds)
	    {
            if (*p & bit)
	       {
               fprintf(stderr,"hdbck: "
		  "byte %d bit %02x "
	          "hash record %d\n",
	          i, bit, bitnum);
	       hshhdr->reason = HASH_ERR + 604;
               return(HASH_ERR + 604);
	       } /* invalid pointer */
	    } /* if in hash area */
	 else if (bitnum <= rcds)
	    {
            if (!(*p & bit))
	       {
               fprintf(stderr,"hdbck: "
		  "byte %d bit %02x "
	          "missing record %d\n",
	          i, bit, bitnum);
	       hshhdr->reason = HASH_ERR + 605;
               return(HASH_ERR + 605);
	       } /* missing overflow */
	    } /* if in ovfl area */
         bit <<= 1;
	 } /* for each bit in byte */
      p++;
      i++;
      } /* for each byte in bitmap */

   return(DBCKGOOD);
   } /* hdbck */
